#include <stdlib.h> 
#include <inttypes.h>
#include <avr/io.h>
#include <avr/interrupt.h>
//#include <avr/sleep.h>
//#include <avr/pgmspace.h>
 #include <avr/eeprom.h>

 #include <math.h>
 
//#include "i2c.h"		// include i2c support
//#include "i2csw.h"		// include i2c support
//#include "a2d.h"		// include A/D converter function library
#include <util/twi.h>		/* Note [1] */

//for avrlib
#include "global.h"

#define PLANES	  8
#define FRAMESIZE 8
//cube is 4*4*4 leds, each led 2 colors: blue and red
//this is arranged as 4*4columnes driven by a tlc5940
//the planes are done by a 74hct138, connected to mosfets

//plane control:
//portc: pc0, pc1, pc2, 
//PD6 can disable all rows

//pc4 = SDA
//pc5 = SCL

#define SAA1064_0 0x7000
#define SAA1064_1 0x7200
#define SAA1064_2 0x7400
#define SAA1064_3 0x7600

int SA1024xx_write(uint16_t eeaddr, int len, uint8_t *buf);
void SetLedInCube( u08 iX, u08 iY, u08 iZ, u08 value);
void ZeroBackBuffer(void);
void ZeroBuffers(void);
u08 GetLedInCube(u08 buffer, u08 iX, u08 iY, u08 iZ );
void SetIndicatorForModeAndSpeed(void);
void GetKeys(void);
void CheckKeyAndSetMode(void);

//fun stuff, the pattern generators
void FullCube(u08 value);
void PlaneOrAlvine(u08 bPlane);
void BlinkFullCube(void);
void FillRandom(void);
void Raindrops(void);
void LifeInCube(void);
//void Sequencer(void);
void Snake(void);
void Cloud(void);

static volatile unsigned char planes[8] = 
{
	0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07
};

volatile unsigned char frontbuf = 0;
volatile unsigned char backbuf = 1;
volatile unsigned char cubedata[2][PLANES][FRAMESIZE];

volatile u08 timerCnt;

// Last character read from the UART.
volatile char rxbuff;

//x,y,z 
static u08 planeLookup[8] = { 0x05, 0x03, 0x06, 0x00, 0x01, 0x07, 0x04, 0x02 };

static u08 ledLookup[8][8] = { 
{ 38,	46,	39,	47,	30,	25,	31,	24 },
{ 36,	44,	37,	45,	41,	34,	40,	35 },
{ 33,	42,	32,	43,	28,	27,	29,	26 },
{  3,	 4,	 2,	 5,	 1,	 6,	 0,	7  },
{ 57,	49,	56,	48,	15,	23,	22,	14 },
{ 59,	51,	58,	50,	13,	21,	12,	20 },
{ 61,	53,	60,	52,	11,	18,	10,	19 },
{ 63,	55,	62,	54,	 9,	17,	 8,	16 }
};


/*
 * Bits that are set inside interrupt routines, and watched outside in
 * the program's main loop.
 */
volatile struct
{
  uint8_t tmr_int: 1;
  uint8_t adc_int: 1;
  uint8_t rx_int: 1;
}
intflags;


void ZeroBuffers(void)
{
	u08 i,j,k;
	for( i=0;i<2;i++ )
		for( j=0;j<PLANES;j++ )
			for( k=0;k<FRAMESIZE;k++ )
				cubedata[i][j][k] = 0x00;
}

void ZeroBackBuffer(void)
{
	u08 j,k;
	for( j=0;j<PLANES;j++ )
		for( k=0;k<FRAMESIZE;k++ )
			cubedata[backbuf][j][k] = 0x00;
}

void SetLedInCube( u08 iX, u08 iY, u08 iZ, u08 value)
{
	//z define plane
	u08 plane = planeLookup[iZ];
		
	//x and y define lednum (0..63)
	u08 lednum = ledLookup[iX][iY];
		
	u08 ind = (lednum>>3) & 0x07;
	u08 shiftvalue = lednum & 0x07;
			
	//uneven lednums:
	if(value!=0){ //turn on
		cubedata[backbuf][plane][ind]  |=  _BV(shiftvalue); //= 0xff;//
	}
	else{ //turn off
		cubedata[backbuf][plane][ind] &= ~(0x01 << shiftvalue); 
	}	
}

u08 GetLedInCube(u08 buffer, u08 iX, u08 iY, u08 iZ )
{
	u08 value = 0;
	
	//z define plane
	u08 plane = planeLookup[iZ];
		
	//x and y define lednum (0..63)
	u08 lednum = ledLookup[iX][iY];
		
	u08 ind = (lednum>>3) & 0x07;
	u08 shiftvalue = lednum & 0x07;
	
	value = (cubedata[buffer][plane][ind] >> shiftvalue) & 0x01;
	
	return value;
}

void InitSAA1064(uint16_t adress)
{
	u08 buffer[3] = { 0x76, 0x00, 0x00 }; //index, setup, 4xleds
	SA1024xx_write( adress, 3, buffer);
}

void SendPlane2Drivers(u08 iPlane)
{	
	u08 buffer[3] = { 0, 0, 0 };
	buffer[0] = cubedata[frontbuf][iPlane][0];
	buffer[1] = cubedata[frontbuf][iPlane][1];	
	SA1024xx_write( SAA1064_0+1, 2, buffer );

	buffer[0] = cubedata[frontbuf][iPlane][2];
	buffer[1] = cubedata[frontbuf][iPlane][3];	
	SA1024xx_write( SAA1064_1+1, 2, buffer );
	
	buffer[0] = cubedata[frontbuf][iPlane][4];
	buffer[1] = cubedata[frontbuf][iPlane][5];	
	SA1024xx_write( SAA1064_2+1, 2, buffer );
	
	buffer[0] = cubedata[frontbuf][iPlane][6];
	buffer[1] = cubedata[frontbuf][iPlane][7];	
	SA1024xx_write( SAA1064_3+1, 2, buffer );
}

volatile u08 bSwap = 0;
void SwapBuffers(void)
{
	bSwap = 1;
	while(bSwap)
		;	
}

volatile u08 planecnt = 0;
volatile u08 timerdiv = 0;
ISR(SIG_OUTPUT_COMPARE1A)
{
//refresh screen routine
	
	//set blank (no row selected)
	cbi( PORTD, PD6 );

	unsigned char tmpbuf = backbuf;	
	if(planecnt == 8){
		planecnt = 0;		
		if(bSwap==1){
			backbuf = frontbuf;
			frontbuf = tmpbuf;					
			bSwap = 0; //swapped
		}
	}

	SendPlane2Drivers(planecnt);

	PORTC &= ~0x07;
	PORTC |= planes[planecnt];
	planecnt++;
		
	// set, to enable plane drivers
	sbi( PORTD, PD6 );	
}

//this happens every 1.024ms
ISR(SIG_OVERFLOW0)
{
	timerdiv++;
	if(timerdiv==10){
		timerdiv=0;
		timerCnt++; //raise number, give a click every 10ms (10perSec) 	
		intflags.tmr_int = 1;	
	}
}






#define TWI_SLA_24CXX	0x00	/* E2 E1 E0 = 0 0 0 */
#define MAX_ITER	200
#define PAGE_SIZE 8
uint8_t twst;

/*
 * Write "len" bytes into EEPROM starting at "eeaddr" from "buf".
 *
 * This is a bit simpler than the previous function since both, the
 * address and the data bytes will be transfered in master transmitter
 * mode, thus no reselection of the device is necessary.  However, the
 * EEPROMs are only capable of writing one "page" simultaneously, so
 * care must be taken to not cross a page boundary within one write
 * cycle.  The amount of data one page consists of varies from
 * manufacturer to manufacturer: some vendors only use 8-byte pages
 * for the smaller devices, and 16-byte pages for the larger devices,
 * while other vendors generally use 16-byte pages.  We thus use the
 * smallest common denominator of 8 bytes per page, declared by the
 * macro PAGE_SIZE above.
 *
 * The function simply returns after writing one page, returning the
 * actual number of data byte written.  It is up to the caller to
 * re-invoke it in order to write further data.
 */
 
int SA1024xx_write(uint16_t eeaddr, int len, uint8_t *buf)
{
  uint8_t sla, n = 0;
  int rv = 0;

  sla = (((eeaddr >> 8) & 0xFF) );

  restart:
  if (n++ >= MAX_ITER)
    return -1;
  begin:

  /* Note [15] */
  TWCR = _BV(TWINT) | _BV(TWSTA) | _BV(TWEN); /* send start condition */
  while ((TWCR & _BV(TWINT)) == 0) ; /* wait for transmission */
  switch ((twst = TW_STATUS))
    {
    case TW_REP_START:		/* OK, but should not happen */
    case TW_START:
      break;

    case TW_MT_ARB_LOST:
      goto begin;

    default:
      return -1;		/* error: not in start condition */
				/* NB: do /not/ send stop condition */
    }

  /* send SLA+W */
  TWDR = sla | TW_WRITE;
  TWCR = _BV(TWINT) | _BV(TWEN); /* clear interrupt to start transmission */
  while ((TWCR & _BV(TWINT)) == 0) ; /* wait for transmission */
  switch ((twst = TW_STATUS))
    {
    case TW_MT_SLA_ACK:
      break;

    case TW_MT_SLA_NACK:	/* nack during select: device busy writing */
      goto restart;

    case TW_MT_ARB_LOST:	/* re-arbitrate */
      goto begin;

    default:
      goto error;		/* must send stop condition */
    }

  TWDR = eeaddr;		/* low 8 bits of addr */
  TWCR = _BV(TWINT) | _BV(TWEN); /* clear interrupt to start transmission */
  while ((TWCR & _BV(TWINT)) == 0) ; /* wait for transmission */
  switch ((twst = TW_STATUS))
    {
    case TW_MT_DATA_ACK:
      break;

    case TW_MT_DATA_NACK:
      goto quit;

    case TW_MT_ARB_LOST:
      goto begin;

    default:
      goto error;		/* must send stop condition */
    }

  for (; len > 0; len--)
    {
      TWDR = *buf++;
      TWCR = _BV(TWINT) | _BV(TWEN); /* start transmission */
      while ((TWCR & _BV(TWINT)) == 0) ; /* wait for transmission */
      switch ((twst = TW_STATUS))
	{
	case TW_MT_DATA_NACK:
	  goto error;		/* device write protected -- Note [16] */

	case TW_MT_DATA_ACK:
	  rv++;
	  break;

	default:
	  goto error;
	}
    }
  quit:
  TWCR = _BV(TWINT) | _BV(TWSTO) | _BV(TWEN); /* send stop condition */

  return rv;

  error:
  rv = -1;
  goto quit;
}

void InitI2C (void)
{

	u08 bitrate_div;
	// set i2c bitrate
	// SCL freq = F_CPU/(16+2*TWBR))
	#ifdef TWPS0
		// for processors with additional bitrate division (mega128)
		// SCL freq = F_CPU/(16+2*TWBR*4^TWPS)
		// set TWPS to zero
		cbi(TWSR, TWPS0);
		cbi(TWSR, TWPS1);
	#endif

	// calculate bitrate division	
	bitrate_div = ((F_CPU/1000l)/120); //200kHz
	if(bitrate_div >= 16)
		bitrate_div = (bitrate_div-16)/2;
	outb(TWBR, bitrate_div);
	
	//setup current(I) values for the columndrivers, make sure they're turned off?
	InitSAA1064( SAA1064_0 );
	InitSAA1064( SAA1064_1 );
	InitSAA1064( SAA1064_2 );
	InitSAA1064( SAA1064_3 );
		
}

void InitIO (void)
{
	ZeroBuffers();

	//plane control pins as output
	DDRC = _BV(PC0) | _BV(PC1) | _BV(PC2);
	PORTC = 0;
	
	//couple of leds on PD7, PB012, plane control on PD6
	DDRD = _BV(PD7) | _BV(PD6);
	DDRB = _BV(PB0) | _BV(PB1) | _BV(PB2);
		
	PORTD = 0xff;
	PORTB = 0x07;
	
	//clear leds for init
	sbi( PORTB, PB0 );
	sbi( PORTB, PB1 );
	sbi( PORTB, PB2 );
	sbi( PORTD, PD7 );

	//make sure plane select is turned off
	//set blank (no row selected)
	cbi( PORTD, PD6 );
		
	//init i2c
	InitI2C();
	
	//Timer0 clocksource:
	TCCR0 = 0x03; //clock/64, a tick every 1.024ms
	
	//0xFFFF gives a 30.5Hz refresh rate: 16MHz/(0xFFFF x 8)
	//0x8235 gives 60Hz

	//timer1 config:
	//interrupt every 0x8235 clocks, so cmp1A
	//with reset of timer1	
	//oc1A,
	OCR1A = 0x8235;
//	OCR1A = 0x7E02; //62Hz no flicker, good brightness
			
	//clear on compare, clock set to mainclock
	TCCR1B = _BV(WGM12) | _BV(CS10);
	
	//set interrupt on compareA, //enable overflow interrupt for timer0
	TIMSK = _BV(OCIE1A) | _BV(TOIE0);
		
	//enable interrupts
    sei ();
	
}


//==============================================================================

void ShortDelay(void )
{
    uint32_t l = 0;
	for(l = 0; l < (F_CPU / 50); ++l);  // delay loop
}


/*void LongDelay(void )
{
    uint32_t l = 0;
	for(l = 0; l < (2*F_CPU); ++l);  // delay loop
}
*/

//=================================================================

//full on
void FullCube(u08 value)
{
	u08 x,y,z;
	for( x=0;x<8;x++ ){
		for( y=0;y<8;y++ ){
			for( z=0;z<8;z++ )	{
				SetLedInCube( x, y, z, value );
			}
		}
	}
}

//=================================================================

//Blink full cube
volatile u08 fullBlink = 0;
void BlinkFullCube(void)
{
	if(fullBlink){
		fullBlink = 0;
		FullCube(fullBlink);
	}
	else{	
		fullBlink = 1;
		FullCube(fullBlink);
	}
}

//=================================================================

//random
volatile u08 rndCnt;

u16 lednums[20];

void GetFromIndex(u08* px, u08* py, u08* pz, u16 index)
{
	//index = z + (8*y) + (64*x);
	*px = (index >> 6) & 0x7;
	*py = (index >> 3) & 0x7;
	*pz = index & 0x7;
}

void FillRandom(void)
{	
	//keep 20 leds on, shift new led in with every step
	//so led stays on for 20 steps
	
	//new random number
	u16 newled = 0;
	u08 bOk = 0;
	u08 stackit;
	while(bOk==0){
		u32 randomNum = random() ;
		newled = randomNum % 0x000001ff; //clip at 511 (0..511)		
		//see if newled is already active
		bOk = 1;
		for( stackit=0; stackit<20; stackit++ ){		
			if( lednums[stackit] ==newled )
				bOk = 0;
		}				
	}
		
	//shift new number in register
	for( stackit=19; stackit>0; stackit-- ){		
		lednums[stackit] = lednums[stackit-1];
	}
	lednums[0] = newled;
			
	ZeroBackBuffer();
	
	u08 x,y,z;
	u16 ind;
	for( ind=0; ind<20; ind++ ){
		GetFromIndex( &x, &y, &z, lednums[ind]);
		SetLedInCube( x, y, z, 1 );
	}	
}

//=================================================================

void ShiftPlanes(u08 direction)
{
//direction: x+, y+, z+, x-, y-, z-
	u08 x,y,z;
	u08 val = 0;
	if(direction>2){
		u08 xMax,yMax,zMax;
		
		//neg dir, x-, y-, z-
		xMax = zMax = yMax = 8;
		if(direction==3)
			xMax--;
		else if(direction==4)
			yMax--;
		else if(direction==5)
			zMax--;
		
		for( x=0; x<xMax; x++ ){
			for( y=0; y<yMax; y++ ){
				for( z=0; z<zMax; z++ ){

					if(direction==3) //x
						val = GetLedInCube( frontbuf, x+1, y,   z );
					else if(direction==4) //y
						val = GetLedInCube( frontbuf, x,   y+1, z );
					else if(direction==5) //z
						val = GetLedInCube( frontbuf, x,   y,   z+1 );

					SetLedInCube( x, y, z, val );
				}
			}
		}	
		
	}
	else{
		u08 xMin, yMin, zMin;
		
		//pos dir, x+, y+, z+
		xMin = zMin = yMin = 0;
		if(direction==0)
			xMin++;
		else if(direction==1)
			yMin++;
		else if(direction==2)
			zMin++;
		
		for( x=8;x>xMin;x-- ){
			for( y=8;y>yMin;y-- ){
				for( z=8;z>zMin;z-- )	{

					if(direction==0) //x
						val = GetLedInCube( frontbuf, x-2, y-1, z-1 );
					else if(direction==1) //y
						val = GetLedInCube( frontbuf, x-1, y-2, z-1 );
					else if(direction==2) //z
						val = GetLedInCube( frontbuf, x-1, y-1, z-2 );

					SetLedInCube( x-1, y-1, z-1, val );
				}
			}
		}	
	}
}

//rain
u08 rainstack[6];
u08 fallenrain[6];
u08 oldfallenrain[6];
void Raindrops(void)
{
	//push random numbers on stack (0, 1, 2, or 3)
	u32 randomNum = random() ;
	
	u08 numleds = randomNum % 0x00000006;
	u08 cntr = 0;
	u08 newled = 0;
	
	for(cntr=0;cntr<numleds;cntr++){
		//new random number
		u08 bOk = 0;
		u08 stackit;
	
		while(bOk==0){
			randomNum = random() ;
			newled = randomNum % 0x0000003f; //clip at 64 (0..63)		
			//see if newled is already active
			bOk = 1;
			for( stackit=0; stackit<6; stackit++ ){		
				if( rainstack[stackit] == newled )
					bOk = 0;
			}				
		}
		
		//push on stack
		for( stackit=5; stackit>0; stackit-- ){		
			rainstack[stackit] = rainstack[stackit-1];
		}
		rainstack[0] = newled;
						
	}
	
	//clear cube
	ZeroBackBuffer();
	
	//shift z plane down (image in frintbuf is shifted down)
	ShiftPlanes(5);	

	//store lower plane in array
	u08 x,y,val,raincnt;
	raincnt = 0;
	for( x=0; x<8; x++ ){
		for( y=0; y<8; y++ ){
			val = GetLedInCube( backbuf, x, y, 0 );
			if(val)
				fallenrain[raincnt++] = (x<<3)|y;									
		}
	}
	
	//use old array to splash
	for( raincnt=0; raincnt<6; raincnt++ ){
		x = (oldfallenrain[raincnt] >> 3) & 0x7;
		y = oldfallenrain[raincnt] & 0x7;

		SetLedInCube( x, y, 0, 1 );
		
		//x+1,x-1,y-1,y+1 (splash effect)		
/*		if(x<7)
			SetLedInCube( x+1, y, 0, 1 );
		if(x>0)
			SetLedInCube( x-1, y, 0, 1 );
		if(y<7)
			SetLedInCube( x, y+1, 0, 1 );	
		if(y>0)
			SetLedInCube( x, y-1, 0, 1 );	*/
	}
	
	//swap arrays
	for( raincnt=0; raincnt<6; raincnt++ ){
		oldfallenrain[raincnt] = fallenrain[raincnt];
	}	
		
	//fill upper plane with new rainstack
	for( cntr=0; cntr<6; cntr++ ){
		u08 x,y;
		x = (rainstack[cntr] >> 3) & 0x7;
		y = rainstack[cntr] & 0x07;
		SetLedInCube( x, y, 7, 1 );	
	}
		
}

#define	R1 3
#define R2 4
#define R3 3
#define R4 2

//=================================================================
u08 GetLife(u08 x, u08 y, u08 z)
{
	//count lighted neighbors (there are 26 neighbors!)
	//we can also make this cyclic, so if i>7 then i=0
	
	u08 xmin, xmax, ymin, ymax, zmin, zmax;
	xmin = ymin = zmin = 0;
	xmax = ymax = zmax = 8;
	
	if(x>0)
		xmin = x - 1;
	if(x<7)
		xmax = x + 2;
	if(y>0)
		ymin = y - 1;
	if(y<7)
		ymax = y + 2;
	if(z>0)
		zmin = z - 1;
	if(z<7)
		zmax = z + 2;
		
	u08 retval = 0;
	u08 ix,iy,iz,val;
	val = 0;
	for( ix=xmin; ix<xmax; ix++ ){
		for( iy=ymin; iy<ymax; iy++ ){
			for( iz=zmin; iz<zmax; iz++ ){
				if( (ix==x)&&(iy==y)&&(iz==z) )
					 retval = 0;
				else{
					val += GetLedInCube( frontbuf, ix, iy, iz );
				}					
			}
		}
	}
/*	
	//cyclic
	if(1){
		if(x==0){
			xmin = 7;
			xmax = 8;
		}
		if(x==7){
			xmin = 0;
			xmax = 1;
		}
		if(y==0){
			ymin = 7;
			ymax = 8;
		}
		if(y==7){
			ymin = 0;
			ymax = 1;
		}
		if(z==0){
			zmin = 7;
			zmax = 8;
		}
		if(z==7){
			zmin = 0;
			zmax = 1;
		}
		
		for( ix=xmin; ix<xmax; ix++ )
			for( iy=ymin; iy<ymax; iy++ )
				for( iz=zmin; iz<zmax; iz++ )
						val += GetLedInCube( frontbuf, ix, iy, iz );
	}
*/		
			
	retval = GetLedInCube( frontbuf, x, y, z );	
	if ( (retval==0)&&(val>=R1)&&(val<=R2) )
		retval = 1;
	else if ( (retval!=0)&&((val>R3)||(val<R4)) ) 
		retval = 0;	
	
	return retval;
}


//life
volatile u08 ilifecnt = 0;
void LifeInCube(void)
{
	//clear cube
	ZeroBackBuffer();

	//see how many leds are currently lighted in frontbuf
	u16 numleds = 0;
	u08 x,y,z;
	for( x=0; x<8; x++ )
		for( y=0; y<8; y++ )
			for( z=0; z<8; z++ )
				if( GetLedInCube( frontbuf, x, y, z ) == 1 )
					numleds++;

	//"life iteration"
	for( x=0; x<8; x++ )
		for( y=0; y<8; y++ )
			for( z=0; z<8; z++ )
				if( GetLife( x, y, z ) == 1 )
					SetLedInCube( x, y, z, 1 );					
	
		
	//if <5 leds were lighted, new random seed of 4 extra for next iteration
	if(numleds<5){
		u08 cnt = 0;
		for( cnt=0; cnt<4; cnt++ ){
			u16 index = 0;
			u32 randomNum = random() ;
			index = randomNum % 0x000001ff; //clip at 64 (0..63)		
			GetFromIndex( &x, &y, &z, index );
			SetLedInCube( x, y, z, 1 );
		}
	}
	
	//every 120 iterations, inject 4 random leds
	ilifecnt++;
	if(ilifecnt>120){
		u08 cnt = 0;
		for( cnt=0; cnt<4; cnt++ ){
			u32 randomNum = random() ;
			u16 index = randomNum % 0x000001ff; //clip at 64 (0..63)		
			GetFromIndex( &x, &y, &z, index );
			SetLedInCube( x, y, z, 1 );
		}
		ilifecnt=0;
	}
	
}

//=================================================================
/*
volatile u16 seqledcnt = 0;
void Sequencer(void)
{
	//clear
	ZeroBackBuffer();

	u16 lednum;
	u08 x,y,z;
	for( x=0;x<8;x++ )
		for( y=0;y<8;y++ )
			for( z=0;z<8;z++ )	{
					lednum = z + (8*y) + (64*x);
					if( lednum<=seqledcnt )//|| (r2==lednum) )
						SetLedInCube( x, y, z, 1 );
			}
		
	seqledcnt++;
	if(seqledcnt>511)
		seqledcnt=0;
}
*/

//=================================================================

#define SNAKE_LENGTH 6
u16 snakes[4][6];
u08 dir[4];

//dir +x,-x,+y,-y,+z,-z
u08 newdirfromold(u08 x, u08 y, u08 z, u08 dir)
{
	u08 bOk;
	u08 newdir = dir;
	u32 randomNum;

	bOk=0;
	while(bOk==0){
		randomNum = random() ;
		newdir = randomNum % 0x00000006;
		//see if this dir is possible				
		bOk = 1;
		if(newdir==0){
			if( (x+1>6) || (dir==0) || (dir==1) )
				bOk = 0;//choose a new direction that doesn't hit a wall, nor is the same or turns back
		}
		if(newdir==1){
			if( (x==0) || (x-1<1) || (dir==0) || (dir==1) )
				bOk = 0;//choose a new direction that doesn't hit a wall, nor is the same or turns back
		}
		if(newdir==2){
			if( (y+1>6) || (dir==2) || (dir==3) )
				bOk = 0;//choose a new direction that doesn't hit a wall, nor is the same or turns back
		}
		else if(newdir==3){
			if( (y==0) || (y-1<1) || (dir==2) || (dir==3) )
				bOk = 0;//choose a new direction that doesn't hit a wall, nor is the same or turns back
		}
		else if(newdir==4){
			if( (z+1>6) || (dir==4) || (dir==5) )
				bOk = 0;//choose a new direction that doesn't hit a wall, nor is the same or turns back
		}
		else if(newdir==5){
			if( (z==0) || (z-1<1) || (dir==4) || (dir==5) )
				bOk = 0;//choose a new direction that doesn't hit a wall, nor is the same or turns back
		}								
	}		
	return newdir;
}

u08 MoveAndGetNewDir( u16* lednum, u08 dir )
{
	u08 x, y, z;
	GetFromIndex( &x, &y, &z, *lednum);
	u08 newdir = dir;
	u32 randomNum;

	//randomize: step or not
	randomNum = random() ;
	u08 dostep = randomNum % 0x00000006;
	
	if(dostep == 0){
		*lednum = z + (8*y) + (64*x);
		return newdir;	
	}

	switch(dir){
		case 0:
			x++;
			if( (x>6) || (dostep == 6) ){
				//choose a new direction that doesn't hit a wall
				//direction can be 2,3,4,5
				newdir = newdirfromold( x, y, z, dir);
			}									
		break;
		case 1:
			x--;
			if( (x<1) || (dostep == 6) ){
				//choose a new direction that doesn't hit a wall
				//direction can be 2,3,4,5
				newdir = newdirfromold( x, y, z, dir);
			}									
		break;
		case 2:
			y++;
			if( (y>6) || (dostep == 6) ){
				//choose a new direction that doesn't hit a wall
				newdir = newdirfromold( x, y, z, dir);
			}									
		break;
		case 3:
			y--;
			if( (y<1) || (dostep == 6) ){
				//choose a new direction that doesn't hit a wall
				newdir = newdirfromold( x, y, z, dir);
			}									
		break;
		case 4:
			z++;
			if( (z>6) || (dostep == 6) ){
				//choose a new direction that doesn't hit a wall
				newdir = newdirfromold( x, y, z, dir);
			}									
		break;
		case 5:
			z--;
			if( (z<1) || (dostep == 6) ){
				//choose a new direction that doesn't hit a wall
				newdir = newdirfromold( x, y, z, dir);
			}									
		break;
		default:
		break;
	}		

	*lednum = z + (8*y) + (64*x);
	return newdir;
}


void Snake(void)
{
	u08 x,y,z,t;//,bOk;
	u08 snake;
	u08 stackit;
	
	//random start value, fill full buffer with this number
	//start value == 0, random dir...
			
	//shift all buffers
	for(snake=0;snake<4;snake++)			
		for( stackit=SNAKE_LENGTH-1; stackit>0; stackit-- )		
			snakes[snake][stackit] = snakes[snake][stackit-1];
				
	//random 1 of 4 directions if next step would be out of a wall or on another snake
	//check if possible dir, if not try again		
	for(snake=0;snake<4;snake++)
		dir[snake] = MoveAndGetNewDir( &(snakes[snake][0]), dir[snake] );	
		
	//clear
	ZeroBackBuffer();
	
	//put snake on the cube
	for(snake=0;snake<4;snake++)			
		for( t=0; t<SNAKE_LENGTH; t++ ){
			GetFromIndex( &x, &y, &z, snakes[snake][t]);
			SetLedInCube( x, y, z, 1 );
		}
	
}


//=================================================================

static u08 maxletter;
static u08 letterFromEE;
//"A","L","V","I","N","E"
EEMEM  u08 letterA_EE[19] = { 19,
11,	12,	18,	21,	26,	29,	34,	37,
41,	42,	43,	44,	45,	46,	49,	54,
57,	62 };

EEMEM u08 letterL_EE[11] = { 11,
10,	18,	26,	34,	42,	50,	58,	59,
60,	61	 };					

EEMEM u08 letterV_EE[17] = { 17,
9,	14,	17,	22,	25,	26,	29,	30,
34,	37,	42,	45,	50,	53,	59,	60 };

EEMEM u08 letterI_EE[19] = { 19,
10,	11,	12,	13,	19,	20,	27,	28,
35,	36,	43,	44,	51,	52,	58,	59,
60,	61	};					

EEMEM u08 letterN_EE[22] = { 22,
9,	13,	17,	18,	21,	25,	26,	27,
29,	33,	35,	37,	41,	43,	44,	45,
49,	52,	53,	57,	61 };			

EEMEM u08 letterE_EE[21] = { 21,
9,	10,	11,	12,	13,	14,	17,	25,
33,	34,	35,	36,	41,	49,	57,	58,
59,	60,	61,	62	};

u08 letterCnt = 0;
u08 planestepperplane = 0;
u08 planestepper = 0;
void PlaneOrAlvine(u08 bPlane)
{
	u08 x,y,z;

	//clear cube
	ZeroBackBuffer();
		
	if(planestepper==0){
		u08 xmin, xmax, ymin, ymax, zmin, zmax;

		xmin = ymin = zmin = 0;
		xmax = ymax = zmax = 8;

		//direction: x+, y+, z+, x-, y-, z-
		
		if(planestepperplane==0){
			xmin = 0;
			xmax = 1;
		}   
		if(planestepperplane==1){
			ymin = 0;
			ymax = 1;
		}   
		if(planestepperplane==2){
			zmin = 0;
			zmax = 1;
		}   
		if(planestepperplane==3){
			xmin = 7;
			xmax = 8;
		}   
		if(planestepperplane==4){
			ymin = 7;
			ymax = 8;
		}   
		if(planestepperplane==5){
			zmin = 7;
			zmax = 8;
		}   
		
		//instead of filling the full plane, create a letter from "ALVINE"
		u08 il; 
		u08 index = 0;
		u08 bOn = 0;
		u08* letteradres = &letterA_EE[0];
		for( x=xmin; x<xmax; x++ ){
			for( y=ymin; y<ymax; y++ ){
				for( z=zmin; z<zmax; z++ ){
					bOn = 1;					
					if(bPlane==0){
						bOn = 0;
						switch(letterCnt){
							case 0:
								maxletter = eeprom_read_byte(&letterA_EE[0]);
								letteradres = &letterA_EE[0];
							break;
							case 1:
								maxletter = eeprom_read_byte(&letterL_EE[0]);
								letteradres = &letterL_EE[0];
							break;
							case 2:
								maxletter = eeprom_read_byte(&letterV_EE[0]);
								letteradres = &letterV_EE[0];
							break;
							case 3:
								maxletter = eeprom_read_byte(&letterI_EE[0]);
								letteradres = &letterI_EE[0];								
							break;
							case 4:
								maxletter = eeprom_read_byte(&letterN_EE[0]);
								letteradres = &letterN_EE[0];
							break;
							case 5:
								maxletter = eeprom_read_byte(&letterE_EE[0]);
								letteradres = &letterE_EE[0];
							break;
						}
						maxletter = eeprom_read_byte(&letterE_EE[0]);
						for(il=1;il<maxletter;il++){
							letterFromEE = eeprom_read_byte(&letteradres[il]);
							if(index==letterFromEE)
								bOn = 1;						
						}
					}

					index++;
					SetLedInCube( x, y, z, bOn );
				}
			}
		}
		
		if(bPlane==0){
			letterCnt++;
			if(letterCnt==6)
				letterCnt=0;
		}
	}
	
	if(planestepper!=0)
		ShiftPlanes(planestepperplane);	
	
	planestepper++;
	if(planestepper==8){
		planestepper=0;
		u32 randomNum = random() ;
		planestepperplane = randomNum % 0x00000006;	
	}
	
}






double dT = 0;
void Cloud(void)
{
	u16 centerled = 0;
	u08 t;
	u08 x,y,z;

	u16 surround[33];

	//wandering dot within cube, with neighbors lighted at random
	u32 randomNum;
		
	dT += 0.05;
	u08 A,B,C;
	A = 4; B = 5; C= 6;
	double dX = 8 * (( sin(A*dT) + 1 ) / 2.);
	double dY = 8 * (( sin(B*dT) + 1 ) / 2.);
	double dZ = 8 * (( sin(C*dT) + 1 ) / 2.);
	
	//move into [0..7] space...
	x = dX; if(x>7) x = 7;
	y = dY; if(y>7) y = 7;
	z = dZ; if(z>7) z = 7;
			
	centerled = z + (8*y) + (64*x);
	for(t=0; t<33; t++)
		surround[t] = centerled;

	//randomize surrounding leds
//	GetFromIndex( &x, &y, &z, centerled);
	u08 x1,y1,z1,t1;
	t1 = 0;
	u08 doOn = 0;
	for(x1=x;x1<x+3;x1++)
		for(y1=y;y1<y+3;y1++)
			for(z1=z;z1<z+3;z1++){
				if( (x1>0) && (y1>0) && (z1>0) 
						&& ( x1<9 ) && ( y1<9 ) && ( z1<9 ) ){
						
					//randomise, 25% chance of lighting led
					randomNum = random() ;
					doOn = randomNum & 0x00000003;
	
					if(doOn == 0)						
						surround[t1++] = (z1-1) + (8*(y1-1)) + (64*(x1-1));
				}				
			}

	//12.5% chance
	randomNum = random() ;
	doOn = randomNum & 0x00000007;
	if((z>1)&&(doOn==0))
		surround[t1++] = (z-2) + (8*(y)) + (64*(x));

	doOn = randomNum & 0x00000038;
	if((z<6)&&(doOn==0))
		surround[t1++] = (z+2) + (8*(y)) + (64*(x));
		
	doOn = randomNum & 0x000001C0;
	if((y>1)&&(doOn==0))
		surround[t1++] = (z) + (8*(y-2)) + (64*(x));

	doOn = randomNum & 0x00000E00;
	if((y<6)&&(doOn==0))
		surround[t1++] = (z) + (8*(y+2)) + (64*(x));

	doOn = randomNum & 0x00007000;
	if((x>1)&&(doOn==0))
		surround[t1++] = (z) + (8*(y)) + (64*(x-2));

	doOn = randomNum & 0x00038000;
	if((x<6)&&(doOn==0))
		surround[t1++] = (z) + (8*(y)) + (64*(x+2));

	//clear
	ZeroBackBuffer();
	
	//put cloud on the cube
	for( t=0; t<33; t++ ){
		GetFromIndex( &x, &y, &z, surround[t]);
		SetLedInCube( x, y, z, 1 );
	}

}


enum
{
MODE_ALL_ON,
MODE_ALL_OFF,
MODE_ALL_BLINK,
MODE_RANDOM,
MODE_RAINDROPS,
MODE_LIFE,
MODE_PLANES,
//MODE_SEQUENCE,
MODE_SNAKE,
MODE_ALVINE,
MODE_CLOUD
} __attribute__((packed)) mode = MODE_CLOUD;

enum
{
SPEED_1,
SPEED_2,
SPEED_3,
SPEED_4,
SPEED_5,
SPEED_6
} __attribute__((packed)) speed = SPEED_3;

volatile u08 ledstate = 0;

void SetIndicatorForModeAndSpeed(void)
{
	sbi( PORTB, PB0 );
	sbi( PORTB, PB1 );
	sbi( PORTB, PB2 );
	sbi( PORTD, PD7 );

	switch(mode){
		case MODE_ALL_ON:
			if(ledstate == 1){
				cbi( PORTB, PB0 );
			}
			else{
				sbi( PORTB, PB0 );
			}
			break;
		case MODE_ALL_OFF:
			if(ledstate == 1){
				cbi( PORTD, PD7 );
			}
			else{
				sbi( PORTD, PD7 );
			}
			break;
		case MODE_ALL_BLINK:
			if(ledstate == 1){
				cbi( PORTB, PB0 );
				cbi( PORTD, PD7 );
			}
			else{
				cbi( PORTB, PB0 );
				cbi( PORTD, PD7 );
			}
			break;
		case MODE_RANDOM:
			if(ledstate == 1){
				cbi( PORTB, PB1 );
			}			
			else{
				sbi( PORTB, PB1 );
			}			
			break;
		case MODE_RAINDROPS:
			if(ledstate == 1){
				cbi( PORTB, PB0 );
				cbi( PORTB, PB1 );
			}			
			else{
				sbi( PORTB, PB0 );
				sbi( PORTB, PB1 );
			}			
			break;
		case MODE_LIFE:
			if(ledstate == 1){
				cbi( PORTB, PB0 );
				cbi( PORTD, PD7 );
				cbi( PORTB, PB1 );
			}			
			else{
				sbi( PORTB, PB0 );
				sbi( PORTD, PD7 );
				sbi( PORTB, PB1 );
			}			
		case MODE_PLANES:
			if(ledstate == 1){
				cbi( PORTB, PB2 );
			}			
			else{
				sbi( PORTB, PB2 );
			}			
			break;
/*		case MODE_SEQUENCE:
			if(ledstate == 1){
				cbi( PORTB, PB0 );
				cbi( PORTB, PB2 );
			}			
			else{
				sbi( PORTB, PB0 );
				sbi( PORTB, PB2 );
			}			
			break;*/
		case MODE_SNAKE:
			if(ledstate == 1){
				cbi( PORTB, PB0 );
				cbi( PORTD, PD7 );
				cbi( PORTB, PB2 );
			}			
			else{
				sbi( PORTB, PB0 );
				sbi( PORTD, PD7 );
				sbi( PORTB, PB2 );
			}			
			break;
		case MODE_ALVINE:
			if(ledstate == 1){
				cbi( PORTB, PB1 );
				cbi( PORTB, PB2 );
			}			
			else{
				sbi( PORTB, PB1 );
				sbi( PORTB, PB2 );
			}			
			break;
		case MODE_CLOUD:
			if(ledstate == 1){
				cbi( PORTB, PB0 );
				cbi( PORTB, PB1 );
				cbi( PORTB, PB2 );
			}			
			else{
				sbi( PORTB, PB0 );
				sbi( PORTB, PB1 );
				sbi( PORTB, PB2 );
			}			
			break;
		
		default:
			break;
	}
}

volatile u08 modeKeyState = 0;
volatile u08 modeKeyStateDebounce = 0;
volatile u08 modeKeyStateOld = 0;
volatile u08 speedKeyState = 0;
volatile u08 speedKeyStateDebounce = 0;
volatile u08 speedKeyStateOld = 0;

void GetKeys(void)
{
	if(	 bit_is_clear(PIND, PD4) )
		modeKeyState = 1;
	else
		modeKeyState = 0;

	if(	 bit_is_clear(PIND, PD3) )
		speedKeyState = 1;
	else
		speedKeyState = 0;
}

void CheckKeyAndSetMode(void)
{	
	//only accept changement fron 1 -> 0, then switch mode and wait for next time of changement
	GetKeys();
	
	//debounce
	while( (modeKeyState != modeKeyStateDebounce) ||  (speedKeyState != speedKeyStateDebounce) ){
		ShortDelay();		
		modeKeyStateDebounce = modeKeyState;
		speedKeyStateDebounce = speedKeyState;		
		GetKeys();
	}

	//mode key down, wasn't down before, so switch mode
	if( modeKeyState == 1 && modeKeyStateOld == 0	){
		switch(mode){
			case MODE_ALL_ON:
				mode = MODE_ALL_OFF;
				break;
			case MODE_ALL_OFF:
				mode = MODE_ALL_BLINK;
				break;
			case MODE_ALL_BLINK:
				mode = MODE_RANDOM;
				break;
			case MODE_RANDOM:
				mode = MODE_RAINDROPS;
				break;
			case MODE_RAINDROPS:
				mode = MODE_LIFE;
				break;
			case MODE_LIFE:
				mode = MODE_PLANES;
				break;
			case MODE_PLANES:
				mode = MODE_SNAKE;
				break;
//			case MODE_SEQUENCE:
	//			mode = MODE_SNAKE;
		//		break;
			case MODE_SNAKE:
				mode = MODE_ALVINE;
				break;
			case MODE_ALVINE:
				mode = MODE_CLOUD;
				break;
			case MODE_CLOUD:
				mode = MODE_ALL_ON;
				break;			
			default:
				break;
		}
			
	}
	modeKeyStateOld = modeKeyState;
	
	//speed key down, steep to next speed level
	if( speedKeyState == 1 && speedKeyStateOld == 0	){
		switch(speed){
			case SPEED_1:
				speed = SPEED_2;
				break;
			case SPEED_2:
				speed = SPEED_3;
				break;
			case SPEED_3:
				speed = SPEED_4;
				break;
			case SPEED_4:
				speed = SPEED_5;
				break;
			case SPEED_5:
				speed = SPEED_6;
				break;
			case SPEED_6:
				speed = SPEED_1;
				break;
			default:
				break;
		}
			
	}
	speedKeyStateOld = speedKeyState;
			
	SetIndicatorForModeAndSpeed();
}

int main (void)
{	
	unsigned char stepSpeed = 0;
    InitIO ();
				
	for(;;){		
		//check switch for mode
		 CheckKeyAndSetMode();
		
		switch(speed){
			case SPEED_1:
				stepSpeed = 2;
				break;
			case SPEED_2:
				stepSpeed = 5;
				break;
			case SPEED_3:
				stepSpeed = 8;
				break;
			case SPEED_4:
				stepSpeed = 12;
				break;
			case SPEED_5:
				stepSpeed = 20;
				break;
			case SPEED_6:
				stepSpeed = 50;
				break;
			default:
				break;
		}
 						
		//fill backbuf and swap to front
		if(timerCnt>stepSpeed){
			timerCnt=0;

			if(ledstate==0)
				ledstate = 1;
			else
				ledstate = 0;
	
			switch(mode){
				case MODE_ALL_ON:
					FullCube(1);
					break;
				case MODE_ALL_BLINK:
					BlinkFullCube();
					break;
				case MODE_RANDOM:
					FillRandom();
					break;
				case MODE_RAINDROPS:
					Raindrops();
					break;
				case MODE_LIFE:
					LifeInCube();
					break;
				case MODE_PLANES:
					PlaneOrAlvine(1);
					break;
//				case MODE_SEQUENCE:
//					Sequencer();
//					break;
				case MODE_SNAKE:
					Snake();
					break;
				case MODE_ALVINE:
					PlaneOrAlvine(0);
					break;
				case MODE_CLOUD:
					Cloud();
					break;
				case MODE_ALL_OFF:
				default:
					FullCube(0);
					break;
			}
			SwapBuffers();			
		}		
	}	

    return (0);
}
